"""
Module to assist in the super-common operation of
wrapping values with prefix and suffix strings.
"""

import re
import six
from options import Options, OptionsClass, Prohibited, Transient
from .util import *
from .quoter import *
from .styleset import *


class XMLQuoter(Quoter):

    """
    A more sophisticated quoter for XML elements that manages tags,
    namespaces, and the idea that some elements may not have contents.
    """

    options = Quoter.options.add(
        tag      = None,
        ns       = None,
        atts     = {},
        attquote = single,
        void     = False,
        prefix   = Prohibited,
        suffix   = Prohibited,
    )

    def __init__(self, *args, **kwargs):
        """
        Create an XMLQuoter
        """
        Quoter.__init__(self)

        # take the atts kwargs for special interpretation
        kw_atts_atts = style_attribs(kwargs.pop('atts', None))

        # update remaining kwargs
        opts = self.options = self.__class__.options.push(kwargs)

        # process flat args
        tagspec, attspec = pad(args, 2)

        # combine style attributes from kwargs and flat args
        # (in reverse order of priority)
        atts = {}
        update_style_dict(atts, kw_atts_atts)
        update_style_dict(atts, kwargs)
        update_style_dict(atts, style_attribs(attspec))
        update_style_dict(atts, style_attribs(tagspec))

        # finally set the object's key values
        opts.tag = atts.pop('_tag', None) or opts.tag
        opts.atts = atts


    def _attstr(self, atts, opts):
        """
        Format an attribute dict. Uses this object's default attribute quoter.
        """
        return ' '.join([''] + ["{0}={1}".format(k, opts.attquote(v)) for k, v in atts.items()])


    def __call__(self, *args, **kwargs):
        """
        Quote a value in X/HTML style, with optional attributes.
        """

        # if not args and not opts.void:
        #    return self.clone(**kwargs)

        # take the atts kwargs for special interpretation
        kw_atts_atts = style_attribs(kwargs.pop('atts', None))

        # update remaining kwargs
        opts = self.options.push(kwargs)

        # process flat args
        if opts.void:
            spec = args[0] if args else ''
            value = None
        else:
            value, spec = pad(args, 2)

        # combine style attributes from opts, kwargs, and flat args
        # (in reverse order of priority)
        atts = {}
        update_style_dict(atts, style_attribs(opts.atts))
        update_style_dict(atts, kwargs)
        update_style_dict(atts, kw_atts_atts)
        update_style_dict(atts, style_attribs(spec))

        # if there is a local tag, let it come in force
        opts.tag = atts.pop('_tag', None) or opts.tag


        # construct the resulting attribute string
        astr = self._attstr(atts, opts) if atts else ''


        pstr, mstr = self._whitespace(opts)
        ns = opts.ns + ':' if opts.ns else ''
        if opts.void or not args:
            parts = [ mstr, '<', ns, opts.tag, astr, '>', mstr ]
        else:
            parts = [ mstr, '<', ns, opts.tag, astr, '>', pstr,
                      stringify(value),
                      pstr, '</', ns, opts.tag, '>', mstr ]
        return self._output(parts, opts)


    # could improve kwargs handling of HTMLQuoter

    # question is, should call attributes overwrite, or add to, object atts?
    # may not be a single answer - eg, in case of class especially

    # This might be case where replace is the primary option, but there's
    # an option to add (or even subtract) - say using a class Add, Plus, Subtract,
    # Minus, Relative, Rel, Delta, etc as an indicator

    # To be a full production XML quoter, might need a slightly more robust way
    # to name XML styles that include namespace names, including some sort of
    # rules for handling hyphens in the names (which cannot be in Python
    # identifiers), and perhaps for understanding namespaces (with terminating
    # colon) as part of tag specification. When a tag is auto-instantiated, it
    # could perhaps have its ns defined as part of its definition, like tag is.


class HTMLQuoter(XMLQuoter):

    """
    A more sophisticated quoter that supports attributes and void elements for HTML.
    """

    options = XMLQuoter.options.add(
        ns       = Prohibited,  # HTML doesn't have namespaces
    )

    def __init__(self, *args, **kwargs):
        XMLQuoter.__init__(self, *args, **kwargs)


# HTML/XML comments need normal, not tag-based, quoters
_markup_comment = Quoter(prefix='<!--', suffix='-->', padding=1)

html = StyleSet(
        factory = HTMLQuoter,
        instant = True,
        immediate = HTMLQuoter('html'))

html.comment = _markup_comment

# Tags that don't take content. Their payload is specified in their
# tag name and attributes, if any.
_SELFCLOSING = """
    br img hr input link meta area base col command embed keygen
    param source track
""".strip().split()

for t in _SELFCLOSING:
    html[t] = HTMLQuoter(t, void=True)


xml = StyleSet(
        factory = XMLQuoter,
        instant = True,
        immediate = XMLQuoter('xml'))

xml.comment = _markup_comment
xml.cdata = Quoter("<![CDATA[", "]]>")
xml.pcdata = Quoter("<![PCDATA[", "]]>")

# TODO: Should xml.pcdata do any auto-quoting of text handed to it?
# TODO: should there be xml.CDATA and xml.PCDATA synonyms





from quoter.styleset import *
from quoter import *
import pytest


def test_basic():

    s = StyleSet(Quoter)
    s.braces = braces

    assert s.braces('this') == '{this}'


def test_setattr():
    s = StyleSet(Quoter)

    s.bspace = braces.but(padding=1)
    assert s.bspace('that') == '{ that }'


def test_getitem():
    s = StyleSet(Quoter)
    s.braces = braces

    assert s['braces'] is braces
    assert s['braces']('curly') == '{curly}'


def test_setitem():
    s = StyleSet(Quoter)
    s['braces'] = braces

    assert s['braces'] is braces
    assert s.braces is braces
    assert s['braces']('curly') == '{curly}'
    assert s.braces('curly') == '{curly}'


def test_getattr_and_definition_not_instant():
    s = StyleSet(Quoter)

    s._define("db", "{{", "}}")
    assert s.db("work") == "{{work}}"

    with pytest.raises(ValueError):
        s.dbspace("{{ ", " }}")("works")

    s._define("wook", prefix="WO", suffix="OK")
    assert s.wook('this') == "WOthisOK"


def test_getattr_factory_instant():

    h = StyleSet(HTMLQuoter, instant=True)
    assert h.pre('this', ".work") == "<pre class='work'>this</pre>"
    assert h.free('open source', ".liberty") == \
        "<free class='liberty'>open source</free>"


def test_immediate():
    q = StyleSet(immediate=brackets)
    assert q("this") == '[this]'

    qnone = StyleSet()
    with pytest.raises(KeyError):
        qnone("this")


def test_repr():
    s = StyleSet()
    s.a = 1
    s['b'] = 2
    assert repr(s) in [ 'StyleSet(a=1, b=2)', 'StyleSet(b=2, a=1)']


def test_examples():
    """
    Test examples from the docs
    """

    colon = quote._define('colon', ':')
    assert colon('this') == quote.colon('this') == ':this:'

    assert quote("super") == "'super'"

    cq = StyleSet(factory=Quoter,
                  immediate=Quoter(':'))

    cq._define("two", Quoter('::'))

    assert cq('this') == ':this:'
    assert cq.two('this') == '::this::'



from quoter import *
import pytest


def test_md_basic():
    assert md.i('this') == '*this*'
    assert md.b('that') == '**that**'
    assert md.a('CNN', 'http://www.cnn.com') == \
                "[CNN](http://www.cnn.com)"


def test_md_h():
    assert md.h('title')    == "# title"
    assert md.h('title', 1) == "# title"
    assert md.h('title', 2) == "## title"
    assert md.h('title', 3) == "### title"
    assert md.h('title', 4) == "#### title"
    assert md.h('title', 5) == "##### title"
    assert md.h('title', 6) == "###### title"

    # test close
    assert md.h('title',    close=True) == "# title #"
    assert md.h('title', 1, close=True) == "# title #"
    assert md.h('title', 2, close=True) == "## title ##"
    assert md.h('title', 3, close=True) == "### title ###"
    assert md.h('title', 4, close=True) == "#### title ####"
    assert md.h('title', 5, close=True) == "##### title #####"
    assert md.h('title', 6, close=True) == "###### title ######"

    # test setext
    assert md.h('title',    setext=True) == "title\n=====\n"
    assert md.h('title', 1, setext=True) == "title\n=====\n"
    assert md.h('title', 2, setext=True) == "title\n-----\n"
    assert md.h('title', 3, setext=True) == "title\n-----\n"
    assert md.h('title', 4, setext=True) == "title\n-----\n"


def test_hN():
    assert md.h1('this') == "# this"
    assert md.h2('this') == "## this"
    assert md.h3('this') == "### this"
    assert md.h4('this') == "#### this"
    assert md.h5('this') == "##### this"
    assert md.h6('this') == "###### this"


def test_md_p():
    assert md.p('this') == 'this'
    assert md.p('this ', md.i('and'), ' that') == 'this *and* that'
    assert md.p('some other stuff') == 'some other stuff'


def test_md_doc():
    paras = [md.p('this ', md.i('and'), ' that'),
             md.p('some other stuff')]
    assert md.doc(paras) == \
           'this *and* that\n\nsome other stuff\n'


def test_md_hr():
    assert md.hr() == "-----"


def test_md_doc():
    assert md.doc([md.p("one"), md.hr(), md.p("two")]) == \
                "one\n\n-----\n\ntwo\n"

# FIXME: hr fails to basic style defn problem




from options import attrs
import sys
from inspect import getmembers, ismethod
from .base import QuoterBase
from .util import listing


class StyleSet(object):

    """
    Container for named styles.
    """

    def __init__(self, factory=None, instant=False, immediate=None,
                       promote=None):
        self._factory = factory
        self._instant = instant
        self._immediate = immediate
        self._styles = {}
        if immediate:
            self._promote_methods(promote)

    def _promote_methods(self, other_methods):
        """
        If this ``StyleSet`` logically stands in for / proxies an immediately
        callable function/object (e.g. ``join``), there are cases when
        calls to this instance's attributes such as ``clone``, ``but``, and
        ``fork`` should actually be calls on the prototype. This method
        sets up that correspondence / forwarding arrangement.

        :param list|str other_methods: List (or whitespace separated string listing)
            of methods to explicitly promote/proxy to the ``self._immediate`` prototype.
        """
        prototype = self._immediate
        methods = set(k for k, v in getmembers(prototype, predicate=ismethod)
                                 if not k.startswith('_'))
        if other_methods:
            methods.update(listing(other_methods))
        for m in methods:
            self.__dict__[m] = getattr(prototype, m)

        # is explicit promotion even required?

    def __getitem__(self, name):
        """
        Support for bracket indexing the style name.
        """
        return self._styles[name]

    def __setitem__(self, name, value):
        """
        Support for bracket indexing the style name when assigning.
        """
        self._styles[name] = value

    def __getattr__(self, name):
        """
        Support for attribute getting the style name. This is the core of
        style auto-generation. If an attribute / style exists, it's just
        handed back. But if it isn't, a closure is generated that, when
        called, will use the factory to create the style, and then immediately
        install the named style.
        """

        style = self._styles.get(name)
        if style:
            return style
        else:
            if self._instant:
                f = self._styles[name] = self._factory(name)
                return f
            else:
                raise ValueError("style {0} not yet defined".format(name))
                # f = self._styles[name] = self._factory(*args, **kwargs)
                # return f

                # FIXME: this one probably requires the linkback strategy
                # TODO: or declare this alternate call style dead

    def __setattr__(self, name, value):
        """
        Support for attribute setting the style name.
        """
        if name.startswith("_"):
            self.__dict__[name] = value
        else:
            self._styles[name] = value

    def _define(self, names, *args, **kwargs):
        """
        Callable version of style defnition.
        Returns value if you want to use it in an assigment.

        :param str names: One or more names (aliases) for the style,
            whitespace separated.
        :param *args: Variable arguments passed to the style factory. If first
            arg is a ``Quoter`` instance, this is your style. Otherwise,
            send all the args and kwargs to the factory.
        :param **kwargs: Keyword arguments passed to the style factory
        :returns: The defined style
        """
        if len(args) == 1 and isinstance(args[0], QuoterBase):
            s = args[0]
        else:
            s = self._factory(*args, **kwargs)
        for name in names.strip().split():
            self._styles[name] = s
        return s

    def __call__(self, *args, **kwargs):
        """
        Handle case styleset is called in immediate mode.
        """
        if self._immediate:
            return self._immediate(*args, **kwargs)
        else:
            raise KeyError("No immediate call specified.")

    def __repr__(self):
        clsname = self.__class__.__name__
        return "{0}({1})".format(clsname, attrs(self._styles))



"""
Module to assist in the super-common operation of
wrapping values with prefix and suffix strings.
"""

import re
import six
from options import Options, OptionsClass, Prohibited, Transient
from .base import QuoterBase
from .util import *
from .styleset import StyleSet


class BadStyleName(ValueError):
    pass


class Quoter(OptionsClass, QuoterBase):

    """
    A quote style. Instantiate it with the style information. Call
    it with a value to quote the value.
    """

    options = Options(
        prefix       = None,
        suffix       = None,
        pair         = Transient,
        sep          = '',
        margin       = 0,
        padding      = 0,
        encoding     = None,
    )

    def __init__(self, *args, **kwargs):
        """
        Create a quoting style.
        """
        opts = self.options = self.__class__.options.push(kwargs)
        self._interpret_args(args)

    def _interpret_pair(self, opts):
        if opts.pair is not Transient:
            opts.prefix, opts.suffix = halfstr(opts.pair)
            opts.pair = Transient

    def _interpret_args(self, args):
        """
        Consume 'flat' *args if present when object is constructed.
        Interpret them, and possibly also options already set.
        """
        opts = self.options
        self._interpret_pair(opts)
        if args:
            used = opts.addflat(args, ['prefix', 'suffix'])
            if 'suffix' not in used:
                opts.suffix = opts.prefix
                # this suffix = prefix behavior appropriate for flat args only

    def _whitespace(self, opts):
        """
        Compute the appropriate margin and padding strings.
        """
        pstr = ' ' * opts.padding if isinstance(opts.padding, int) else opts.padding
        mstr = ' ' * opts.margin  if isinstance(opts.margin, int)  else opts.margin
        return (pstr, mstr)

        # could extend the padding and margins with tuples to enable
        # asymmetric before/after settings

    def _output(self, parts, opts):
        """
        Given a list of string parts, concatentate them and output
        with the given encoding (if any).
        """
        outstr = ''.join(parts)
        return outstr.encode(opts.encoding) if opts.encoding else outstr

    def __call__(self, *args, **kwargs):
        """
        Quote the value, according to the current options.
        """
        opts = self.options.push(kwargs)
        self._interpret_pair(opts)
        pstr, mstr = self._whitespace(opts)
        sval = opts.sep.join(stringify(a) for a in args)
        prefix = opts.prefix or ''
        suffix = opts.suffix or ''
        parts = [ mstr, prefix, pstr, sval, pstr, suffix, mstr ]
        return self._output(parts, opts)

    def clone(self, **kwargs):
        """
        Create a new instance whose options are chained to this instance's
        options (and thence to self.__class__.options). kwargs become the
        cloned instance's overlay options.
        """
        cloned = self.__class__()
        cloned.options = self.options.push(kwargs)
        cloned._interpret_pair(cloned.options)
        return cloned

        # NB clone takes only kwargs, not flat args, contra constructor

    but = clone


# create some default named styles

quote = StyleSet(factory=Quoter,
                 instant=False,
                 immediate=Quoter("'"))


braces   = quote._define("braces",   pair='{}')
brackets = quote._define("brackets", pair='[]')
angles   = quote._define("angles",   pair='<>')
parens   = quote._define("parens",   pair='()')
qs = single = quote._define("qs single", "'")
qd = double = quote._define("qd double", '"')
qt = triple = quote._define("qt triple", '"""')
qb = backticks = quote._define("qb backticks", "`")
qdb = doublebackticks = quote._define("qdb doublebackticks", "``")

# and some Unicode styles
anglequote = guillemet = quote._define("anglequote guillemet",
                                           pair=six.u('\u00ab\u00bb'))
chevron = quote._define("chevron",         pair=six.u('\u2039\u203a'))
curlysingle = quote._define("curlysingle", pair=six.u('\u2018\u2019'))
curlydouble = quote._define("curlydouble", pair=six.u('\u201c\u201d'))


class LambdaQuoter(Quoter):

    """
    A Quoter that uses code to decide what quotes to use, based on the value.
    """

    options = Quoter.options.add(
        func   = None,
        prefix = Prohibited,
        suffix = Prohibited,
        pair   = Prohibited,
    )

    def _interpret_args(self, args):
        """
        Consume 'flat' *args if present when object is constructed.
        Interpret them, and possibly also options already set.
        """
        if args:
            self.options.addflat(args, ['func'])

    def __call__(self, value, **kwargs):
        """
        Quote the value, based on the instance's function.
        """
        opts = self.options.push(kwargs)
        pstr, mstr = self._whitespace(opts)
        prefix, value, suffix = opts.func(value)
        parts = [mstr, prefix, pstr, stringify(value), pstr, suffix, mstr]
        return self._output(parts, opts)

        # TODO: Determine if LambdaQuoters can take multiple arguments

lambdaq = StyleSet(
            factory=LambdaQuoter,
            instant=False,
            immediate=LambdaQuoter(lambda v: ('', 'ALL YOUR BASE ARE BELONG TO US', '')))



from quoter import *
import pytest


def test_braces():
    assert braces('this') == '{this}'
    assert braces('this', padding=1) == '{ this }'
    assert braces('this', margin=1) == ' {this} '
    assert braces('this', padding=1, margin=1) == ' { this } '


def test_brackets():
    assert brackets('this') == '[this]'
    assert brackets('this', padding=1) == '[ this ]'
    assert brackets('this', margin=1) == ' [this] '
    assert brackets('this', padding=1, margin=1) == ' [ this ] '


def test_other_named_ascii_styles():
    assert angles('this') == '<this>'
    assert parens('this') == '(this)'
    assert qs('this') == single('this') == "'this'"
    assert qd('this') == double('this') == '"this"'
    assert qt('this') == triple('this') == '"""this"""'
    assert qb('this') == backticks('this') == '`this`'
    assert qdb('this') == doublebackticks('this') == '``this``'


def test_named_unicode_styles():
    assert anglequote('this') == guillemet('this') == six.u('\u00abthis\u00bb')
    assert chevron('this') == six.u('\u2039this\u203a')
    assert curlysingle('this') == six.u('\u2018this\u2019')
    assert curlydouble('this') == six.u('\u201cthis\u201d')


def test_clone():
    """
    Make sure clones properly registered just like normally defined
    styles, that clones are different from original, that they yield
    the expected results, and that after cloning, the originals
    behave exactly as before
    """
    bar2 = quote.bar2 = brackets.clone(margin=2)
    assert bar2 is not brackets
    assert bar2('this') == '  [this]  '
    assert brackets('this') == '[this]'
    assert quote.bar2('this') == '  [this]  '

    bb = html.b.clone(padding=1, margin=2)
    assert bb is not html.b
    assert isinstance(html.b, HTMLQuoter)
    assert isinstance(bb, HTMLQuoter)
    assert bb('this') == '  <b> this </b>  '
    assert html.b('this') == '<b>this</b>'

    sp = html.b.but(atts='span')
    assert sp is not html.b
    assert sp('something') == '<span>something</span>'
    assert sp('else', '.other') == "<span class='other'>else</span>"


def test_but():
    """
    Test the but alias for clone.
    """
    bar2 = brackets.but(margin=2)
    assert bar2 is not brackets
    assert bar2('this') == '  [this]  '
    assert brackets('this') == '[this]'

    bb = html.b.but(padding=1, margin=2)
    assert bb is not html.b
    assert bb('this') == '  <b> this </b>  '
    assert html.b('this') == '<b>this</b>'


def test_set():
    bar2 = quote.bar2 = brackets.clone(margin=2)
    assert bar2('this') == '  [this]  '
    assert quote.bar2('this') == '  [this]  '
    # check that original object is not changed
    assert brackets('this') == '[this]'
    assert quote.brackets('this') == '[this]'

    # check setting options
    bar2.set(margin=1, padding=1)
    assert bar2('this') == ' [ this ] '
    assert quote.bar2('this') == ' [ this ] '
    # check that original object is not changed
    assert brackets('this') == '[this]'
    assert quote.brackets('this') == '[this]'

    # check setting options
    bar2.set(prefix='||', padding=1)
    assert bar2('this') == ' || this ] '
    assert quote.bar2('this') == ' || this ] '
    # check that original object is not changed
    assert brackets('this') == '[this]'
    assert quote.brackets('this') == '[this]'

def test_set_example():
    bars = Quoter('|')
    assert bars('x') == '|x|'
    bars.set(prefix='||', suffix='||')
    assert bars('x') == '||x||'
    bars.set(padding=1)
    assert bars('x') == '|| x ||'

    bart = bars.clone(prefix=']', suffix = '[')
    assert bart('x') == '] x ['

    bartwide = bart.but(margin=2)
    assert bartwide('x') == '  ] x [  '


def test_pairs():
    percent = Quoter(pair='%%')
    assert percent('something') == '%something%'
    doublea = Quoter(pair='<<>>')
    assert doublea('AAA') == '<<AAA>>'

    # From a quoter point of view, the chars kwarg is arguably gilding the
    # lily. But for exercising and exploring flexible parameter handling and
    # the underlying options package, it is quite interesting. It requires a
    # mapping of one user-level argument into two different underlying
    # parameters. It's a good edge-case for testing.

    # TODO: Move pairs= processing into standard magical argument handling scheme


def test_broad_interpretation_of_pairs():
    percent = Quoter(pair='%%')
    assert percent('this', pair='@@') == '@this@'
    dollars = percent.but(pair='$$')
    assert dollars('this') == '$this$'


def test_auto_stringification():
    assert brackets(12) == '[12]'
    assert braces(4.4) == '{4.4}'
    assert double(None) == '"None"'
    assert single(False) == "'False'"


def test_multiple_values():
    assert braces('a', 'b', 'c') == '{abc}'
    assert single('a', 'b', 'c', sep='|') == "'a|b|c'"


def test_shortcuts():
    assert ' '.join([qs('one'), qd('two'), qt('three'), qb('and'), qdb('four')]) == \
        "'one' \"two\" \"\"\"three\"\"\" `and` ``four``"


def test_instant():
    assert Quoter('+[ ', ' ]+')('castle') == '+[ castle ]+'


def test_lambda():
    f = lambda v: ('(', abs(v), ')') if v < 0 else ('', v, '')
    financial = LambdaQuoter(f)
    assert financial(-10) == '(10)'
    assert financial(44) == '44'

    password = LambdaQuoter(lambda v: ('', 'x' * len(v), ''))
    assert password('secret!') == 'xxxxxxx'

    wf = lambda v:  ('**', v, '**') if v < 0 else ('', v, '')
    warning = lambdaq.warning = LambdaQuoter(wf)
    assert warning(12) == '12'
    assert warning(-99) == '**-99**'
    assert warning(-99, padding=1) == '** -99 **'

    assert lambdaq.warning(12) == '12'
    assert lambdaq.warning(-99) == '**-99**'
    assert lambdaq.warning(-99, padding=1) == '** -99 **'


def test_examples():
    assert single('this') == "'this'"
    assert double('that') == '"that"'
    assert backticks('ls -l') == "`ls -l`"
    assert braces('curlycue') == "{curlycue}"
    assert braces('curlysue', padding=1) == '{ curlysue }'

    bars = Quoter('|')
    assert bars('x') == '|x|'

    plus = Quoter('+', '')
    assert plus('x') == '+x'

    variable = Quoter('${', '}')
    assert variable('x') == '${x}'

    vars = Quoter(prefix='${', suffix='}')
    assert vars('y') == '${y}'

    onetwo = Quoter(pair="1221")
    assert onetwo('this') == '12this21'


def test_attribute_invocations():
    assert single('something') == quote.single('something')
    assert single('something', margin=2, padding=3) == \
           quote.single('something', margin=2, padding=3)
    assert braces('b') == quote.braces('b')

    # now test wholesale
    names = ('braces brackets angles parens qs qd qt qb single double triple '
             'backticks anglequote guillemet chevron curlysingle curlydouble')
    for name in names.split():
        main = eval(name)
        attr = eval('quote.' + name)
        assert main is attr
        assert main('string') == attr('string')


def test_redef():
    braces = Quoter('{', '}', padding=1)
    assert braces('this') == '{ this }'
    assert braces('this', padding=0) == '{this}'
